#include "custom.h"
#include <string.h>
#include <stdio.h>

#include <malloc.h>

namespace CustomMatrices {

static void ReadMatrice(uint8_t *matrices, const char *coeffs, int nCoefficients)
{
   if ( !coeffs )
      return;

   for ( int i = 0; i < nCoefficients - 1; i++ )
   {
      char number[256];
      
      strncpy(number, coeffs, strchr(coeffs, ' ') - coeffs);
      number[strchr(coeffs, ' ') - coeffs] = 0;
      matrices[i] = atoi(number);
      if ( matrices[i] == 0 ) matrices[i] = 1;
      coeffs = strchr(coeffs, ' ') + 1;
   }
   matrices[nCoefficients - 1] = atoi(coeffs);
   if ( matrices[nCoefficients - 1] == 0 ) matrices[nCoefficients - 1] = 1;
}

Filter::Filter(PClip child, const Configuration &configuration, IScriptEnvironment *env) :
GenericVideoFilter(child), configuration(configuration), short_term(1000, vi.fps_numerator), long_term(10000, vi.fps_numerator)
{
   nWidth = vi.width;
   nHeight = vi.height;

   /* initialize x264 librairy */
   x264_param_t params;

   x264_param_default( &params );

   /* fills in the custom matrices */
   ReadMatrice(params.cqm_4iy, configuration.pszMatrices[0], 16);
   ReadMatrice(params.cqm_4ic, configuration.pszMatrices[1], 16);
   ReadMatrice(params.cqm_8iy, configuration.pszMatrices[2], 64);
   ReadMatrice(params.cqm_4py, configuration.pszMatrices[3], 16);
   ReadMatrice(params.cqm_4pc, configuration.pszMatrices[4], 16);
   ReadMatrice(params.cqm_8py, configuration.pszMatrices[5], 64);

   params.psz_cqm_file = strdup(configuration.pszMatricesFile);

   /* fills the rest of the params */
   params.i_width = nWidth;
   params.i_height = nHeight;
   params.i_csp = X264_CSP_I420;
   params.i_frame_total = 0;
   params.i_fps_num = vi.fps_numerator;
   params.i_fps_den = vi.fps_denominator;

   params.analyse.b_psnr = configuration.showMetrics;
   params.analyse.b_ssim = configuration.showMetrics;

   params.i_keyint_min = 1;
   if ( !configuration.isInter )
   {
      params.i_keyint_max = 1; /* only iframes */
      params.i_scenecut_threshold = -1;
   }
   else
   {
      params.i_keyint_max = 250;
      params.i_keyint_min = 25;
      params.i_scenecut_threshold = 40;
   }

   params.b_deblocking_filter = configuration.isDeblocking;
   params.i_deblocking_filter_alphac0 = configuration.nDeblockingStrength;
   params.i_deblocking_filter_beta = configuration.nDeblockingStrength;

   params.i_cqm_preset = X264_CQM_CUSTOM;

   params.rc.i_qp_constant = configuration.nQuantizer;
   params.rc.i_bitrate = configuration.nBitrate;

   params.rc.i_aq_mode = configuration.nAq;
   params.rc.f_aq_strength = configuration.fAq;

   if ( !configuration.isInter )
      params.rc.f_ip_factor = 1.0f;
   else
      params.rc.f_ip_factor = 1.4f;

   params.rc.i_qp_min = 1;
   params.rc.i_rc_method = configuration.isCbr ? X264_RC_ABR : X264_RC_CQP;

   params.analyse.intra = X264_ANALYSE_I4x4;
   if ( configuration.is8x8 )
      params.analyse.intra |= X264_ANALYSE_I8x8;
   params.analyse.b_transform_8x8 = configuration.is8x8;

   if ( configuration.isInter )
   {
      params.analyse.inter = X264_ANALYSE_I4x4 | X264_ANALYSE_PSUB16x16 | X264_ANALYSE_PSUB8x8;
      if ( configuration.is8x8 )
         params.analyse.inter |= X264_ANALYSE_I8x8;

      params.analyse.i_me_method = X264_ME_DIA;
      params.analyse.i_me_range = 16;
      params.analyse.i_subpel_refine = configuration.nMe;
      params.analyse.i_mv_range = 512;
   }

   x264 = x264_encoder_open( &params );

   if ( !x264 )
      env->ThrowError("AVCMatrices : unvalid settings");

   x264_picture_alloc( &input, X264_CSP_I420, nWidth, nHeight );

   nFrames = 0;
}

Filter::~Filter()
{
   x264_picture_clean( &input );
   x264_encoder_close( x264 );
}

static void ConvertToX264Picture(const PVideoFrame &frame, int nWidth, int nHeight, x264_picture_t &picture, IScriptEnvironment* env)
{
   env->BitBlt( picture.img.plane[0], picture.img.i_stride[0],
      frame->GetReadPtr(PLANAR_Y), frame->GetPitch(PLANAR_Y), nWidth, nHeight);
   env->BitBlt( picture.img.plane[1], picture.img.i_stride[1],
      frame->GetReadPtr(PLANAR_U), frame->GetPitch(PLANAR_U), nWidth >> 1, nHeight >> 1);
   env->BitBlt( picture.img.plane[2], picture.img.i_stride[2],
      frame->GetReadPtr(PLANAR_V), frame->GetPitch(PLANAR_V), nWidth >> 1, nHeight >> 1);
}

static void ConvertToVideoFrame(PVideoFrame &frame, int nWidth, int nHeight, x264_picture_t &picture, IScriptEnvironment* env)
{
   env->BitBlt( frame->GetWritePtr(PLANAR_Y), frame->GetPitch(PLANAR_Y),
      picture.img.plane[0], picture.img.i_stride[0], nWidth, nHeight);
   env->BitBlt( frame->GetWritePtr(PLANAR_U), frame->GetPitch(PLANAR_U),
      picture.img.plane[1], picture.img.i_stride[1], nWidth >> 1, nHeight >> 1);
   env->BitBlt( frame->GetWritePtr(PLANAR_V), frame->GetPitch(PLANAR_V),
      picture.img.plane[2], picture.img.i_stride[2], nWidth >> 1, nHeight >> 1);
}

#include "info.h"

PVideoFrame __stdcall Filter::GetFrame(int n, IScriptEnvironment* env)
{
   PVideoFrame dst = env->NewVideoFrame(vi);
   PVideoFrame src = child->GetFrame(n, env);

   x264_nal_t *pNals;
   int nNals;

   ConvertToX264Picture( src, nWidth, nHeight, input, env );
   input.i_pts = nFrames * vi.fps_denominator;
   nFrames++;

   x264_encoder_encode( x264, &pNals, &nNals, &input, &output );

   ConvertToVideoFrame( dst, nWidth, nHeight, output, env );

   int nSize = 0;
   for ( int i = 0; i < nNals; i++ )
      nSize += pNals[i].i_payload;

   short_term.push( nSize, input.i_pts );
   long_term.push( nSize, input.i_pts );

   if ( configuration.showMetrics || configuration.showSize )
   {
      char typesize[100];
      char typepsnr[100];
      int y = 0;
      sprintf(typepsnr, "type : %s", output.i_type == X264_TYPE_P ? "P" : "I");
      DrawString(dst, 0, y++, typepsnr);
      if ( configuration.showMetrics )
      {
         sprintf(typepsnr, "psnr luma : %7.4f, chroma : %7.4f", output.f_psnr[0], output.f_psnr[1]);
         DrawString(dst, 0, y++, typepsnr);
         sprintf(typepsnr, "psnr frame : %7.4f, ssim : %7.6f", output.f_psnr[2], output.f_ssim);
         DrawString(dst, 0, y++, typepsnr);
      }
      if ( configuration.showSize )
      {
         sprintf(typesize, "quantizer : %i, size : %i kbps,", output.i_qpplus1 - 1, nSize * 8 * vi.fps_numerator / (vi.fps_denominator * 1000) );
         DrawString(dst, 0, y++, typesize);
         sprintf(typesize, "bitrate local %i kbps, global : %i kbps", short_term.bitrate() / 1000, long_term.bitrate() / 1000);
         DrawString(dst, 0, y++, typesize);
      }
   }

   return dst;
}

}

AVSValue __cdecl CreateCustomMatrices(AVSValue args, void *user_data, IScriptEnvironment* env)
{
   CustomMatrices::Configuration configuration;

   memset( &configuration, 0, sizeof( configuration ) );

#define CQM_CLIP        0
#define CQM_FILE        1
#define CQM_IL_4x4      2
#define CQM_IC_4x4      3
#define CQM_IL_8x8      4
#define CQM_PL_4x4      5
#define CQM_PC_4x4      6
#define CQM_PL_8x8      7
#define CQM_INTRA       8
#define CQM_QUANTIZER   9
#define CQM_BITRATE     10
#define CQM_DEBLOCKING  11
#define CQM_NAQ         12
#define CQM_FAQ         13
#define CQM_SHOWSIZE    14
#define CQM_SHOWPSNR    15

#define IS_DEF(x) args[CQM_##x].Defined()
#define IVALUE(x,y) args[CQM_##x].AsInt(y)
#define BVALUE(x,y) args[CQM_##x].AsBool(y)
#define FVALUE(x,y) args[CQM_##x].AsFloat(y)
#define SVALUE(x) (IS_DEF(x) ? args[CQM_##x].AsString("") : NULL)

   configuration.is8x8                 = IS_DEF(IL_8x8) || IS_DEF(PL_8x8) || IS_DEF(FILE);
   configuration.isBidir               = false;
   configuration.isCbr                 = IS_DEF(BITRATE);
   configuration.isDeblocking          = IS_DEF(DEBLOCKING);
   configuration.isInter               = !BVALUE(INTRA, false);
   configuration.nBitrate              = IVALUE(BITRATE, 1000);
   configuration.nDeblockingStrength   = IVALUE(DEBLOCKING, 0);
   configuration.nMe                   = 3;
   configuration.nQuantizer            = IVALUE(QUANTIZER, 26);
   configuration.pszMatrices[0]        = SVALUE(IL_4x4);
   configuration.pszMatrices[1]        = SVALUE(IC_4x4);
   configuration.pszMatrices[2]        = SVALUE(IL_8x8);
   configuration.pszMatrices[3]        = SVALUE(PL_4x4);
   configuration.pszMatrices[4]        = SVALUE(PC_4x4);
   configuration.pszMatrices[5]        = SVALUE(PL_8x8);
   configuration.pszMatricesFile       = SVALUE(FILE);
   configuration.showMetrics           = BVALUE(SHOWPSNR, true);
   configuration.showSize              = BVALUE(SHOWSIZE, true);
   configuration.nAq                   = IVALUE(NAQ, 0);
   configuration.fAq                   = FVALUE(FAQ, 1.0);

   return new CustomMatrices::Filter(args[0].AsClip(), configuration, env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
   env->AddFunction("AVCMatrices", "c[file]s[il4]s[ic4]s[il8]s[pl4]s[pc4]s[pl8]s[intra]b[q]i[br]i[dbk]i[aq]i[aqs]f[size]b[psnr]b", CreateCustomMatrices, NULL);

   return("AVCMatrices");
}


